from neupy.core.config import Configurable
from neupy.algorithms.backprop import WEIGHT_UPDATE


__all__ = ('WeightUpdateConfigurable',)


class WeightUpdateConfigurable(Configurable):
    optimization_type = WEIGHT_UPDATE



from numpy import where
from numpy.core.umath_tests import inner1d


__all__ = ('sign2bin', 'bin2sign', 'hopfield_energy')


def sign2bin(matrix):
    return where(matrix == 1, 1, 0)


def bin2sign(matrix):
    return where(matrix == 0, -1, 1)


def hopfield_energy(weight, input_data, output_data):
    return -0.5 * inner1d(input_data.dot(weight), output_data)



from numpy import any as np_any

from neupy.core.base import BaseSkeleton
from neupy.core.properties import ChoiceProperty, NonNegativeIntProperty
from neupy.core.config import Configurable


__all__ = ('DiscreteMemory',)


class DiscreteMemory(BaseSkeleton, Configurable):
    """ Base class for discrete memory networks.

    Notes
    -----
    * {discrete_data_note}
    """
    __discrete_data_note = """ Input and output data must contains only \
    binary values.
    """

    __discrete_params = """mode : {'sync', 'async'}
        Indentify pattern recovery mode. ``sync`` mode try recovery a pattern
        using the all input vector. ``async`` mode randomly chose some
        values from the input vector and repeat this procedure the number
        of times a given variable ``n_times``. Defaults to ``sync``.
    n_times : int
        Available only in ``async`` mode. Identify number of random trials.
        Defaults to ``100``.
    """

    shared_docs = {
        'discrete_data_note': __discrete_data_note,
        'discrete_params': __discrete_params
    }

    mode = ChoiceProperty(default='sync', choices=['async', 'sync'])
    n_times = NonNegativeIntProperty(default=100)

    def __init__(self, **options):
        super(DiscreteMemory, self).__init__(**options)
        self.weight = None

        if 'n_times' in options and self.mode != 'async':
            self.logs.warning("You can use `n_times` property only in "
                              "`async` mode.")

    def discrete_validation(self, matrix):
        """ Validate discrete matrix.

        Parameters
        ----------
        matrix : array-like
            Matrix for validation.

        Returns
        -------
        bool
            Got ``True`` all ``matrix`` discrete values are in
            `discrete_values` list and `False` otherwise.
        """
        if np_any((matrix != 0) & (matrix != 1)):
            raise ValueError("This network is descrete. This mean that you "
                             "can use data which contains 0 and 1 values")



from numpy import arccos, dot, reshape
from numpy.linalg import norm

from neupy.core.properties import DictProperty
from neupy.functions import get_partial_for_func
from neupy.functions import (linear, sigmoid, step, tanh, rectifier,
                             softplus, softmax)
from neupy.layers.base import BaseLayer


__all__ = ('Layer', 'LinearLayer', 'SigmoidLayer', 'StepLayer', 'TanhLayer',
           'RectifierLayer', 'SoftplusLayer', 'SoftmaxLayer',
           'EuclideDistanceLayer', 'AngleDistanceLayer')


class Layer(BaseLayer):
    """ The base class of neural network layers.

    Parameters
    ----------
    function_coef : dict
        The default settings for the activation function.
    {layer_params}
    """

    function_coef = DictProperty()

    def __init__(self, *args, **kwargs):
        super(Layer, self).__init__(*args, **kwargs)

        if self.function_coef is not None:
            partial_func = get_partial_for_func(self.activation_function)
            self.activation_function = partial_func(
                self.activation_function, **self.function_coef
            )


class LinearLayer(Layer):
    """ The layer with the linear activation function.

    Parameters
    ----------
    {layer_params}
    """
    activation_function = linear


class SigmoidLayer(Layer):
    """ The layer with the sigmoid activation function.

    Parameters
    ----------
    function_coef : dict
        The default configurations for the sigmoid activation function.
        There is one available parameter ``alpha`` (defaults to ``1``).
        Parameter ``alpha`` controls function shape.
    {layer_params}
    """
    function_coef = DictProperty(default={'alpha': 1})
    activation_function = sigmoid


class StepLayer(Layer):
    """ The layer with the the step activation function.

    Parameters
    ----------
    {layer_params}
    """
    activation_function = step


class TanhLayer(Layer):
    """ The layer with the `tanh` activation function.

    Parameters
    ----------
    function_coef : dict
        The default configurations for the sigmoid activation function.
        There is one available parameter ``alpha`` (defaults to ``1``).
        Parameter `alpha` controls function shape.
    {layer_params}
    """
    function_coef = DictProperty(default={'alpha': 1})
    activation_function = tanh


class RectifierLayer(Layer):
    """ The layer with the rectifier activation function.

    Parameters
    ----------
    {layer_params}
    """
    activation_function = rectifier


class SoftplusLayer(Layer):
    """ The layer with the softplus activation function.

    Parameters
    ----------
    {layer_params}
    """
    activation_function = softplus


class SoftmaxLayer(Layer):
    """ The layer with the softmax activation function.

    Parameters
    ----------
    function_coef : dict
        The default configurations for the softmax activation function.
        There is one available parameter ``temp`` (defaults to ``1``).
        Lower ``temp`` value will make your winner probability closer
        to ``1``. Higher ``temp`` value will make all probabilities
        values equal to each other.
    {layer_params}
    """
    function_coef = DictProperty(default={'temp': 1})
    activation_function = softmax


class EuclideDistanceLayer(BaseLayer):
    """ The layer compute Euclide distance between the input
    value and weights.

    Parameters
    ----------
    {layer_params}
    """
    def output(self, input_value):
        input_data = self.preformat_input(input_value)
        distance = norm(input_data.T - self.weight, axis=0)
        return -reshape(distance, (1, self.weight.shape[1]))


class AngleDistanceLayer(BaseLayer):
    """ The layer compute cosine distance between the input value
    and weights.

    Parameters
    ----------
    {layer_params}
    """
    def output(self, input_value):
        input_data = self.preformat_input(input_value)
        norm_prod = norm(input_data) * norm(self.weight, axis=0)
        summated_data = dot(input_data, self.weight)
        return -reshape(arccos(summated_data / norm_prod),
                        (1, self.weight.shape[1]))



from abc import abstractmethod

from neupy.helpers import preformat_value
from neupy.helpers.logs import Verbose
from .config import ConfigurableWithABC


__all__ = ('BaseSkeleton',)


class BaseSkeleton(ConfigurableWithABC, Verbose):
    """ Base class for all algorithms and networks.
    """
    def get_params(self, deep=False):
        options = {}
        for name, option in self.options.items():
            value = getattr(self, name)
            # Default value is not always valid type. For this reason we
            # need to ignore all the values that have the same value as
            # in default attibute.
            if value != option.value.default:
                options[name] = value
        return options

    def set_params(self, **params):
        self.__dict__.update(params)
        return self

    @abstractmethod
    def train(self, input_data, target_data):
        pass

    @abstractmethod
    def predict(self, input_data):
        pass

    def fit(self, X, y, *args, **kwargs):
        self.train(X, y, *args, **kwargs)
        return self

    def _repr_options(self):
        options = []
        for option_name in self.options:
            option_value = getattr(self, option_name)
            option_value = preformat_value(option_value)

            option_repr = "{}={}".format(option_name, option_value)
            options.append(option_repr)

        return ', '.join(options)

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            self._repr_options()
        )



__all__ = ("docs",)


docs = {
    # ------------------------------------ #
    #                Methods               #
    # ------------------------------------ #

    "last_error": """last_error()
        Returns the last error network result after training procedure
        or ``None`` value if you try to get it before network training.
    """,
    "plot_errors": """plot_errors(logx=False)
        Draws the error rate update plot. It always shows network
        learning progress. When you add cross validation data set
        into training function it displays validation data set error as
        separated curve. If parameter ``logx`` is equal to the
        ``True`` value it displays x-axis in logarithmic scale.
    """,
    "predict": """predict(input_data)
        Predict value.
    """,
    "fit": """fit(\*args, \*\*kwargs)
        The same as ``train`` method.
    """,

    # ------------------------------------ #
    #             Train Methods            #
    # ------------------------------------ #

    "supervised_train": """train(input_train, target_train, input_test=None,\
    target_test=None, epochs=100, epsilon=None):
        Trains network. You can control network training procedure
        iterations with the number of epochs or converge value epsilon.
        Also you can specify ``input_test`` and ``target_test`` and control
        your validation data error on each iteration.
    """,
    "supervised_train_epochs": """train(input_data, target_data, epochs=100):
        Trains network with fixed number of epochs.
    """,
    "unsupervised_train_epochs": """train(input_train, epochs=100):
        Trains network with fixed number of epochs.
    """,
    "unsupervised_train_epsilon": """train(input_train, epsilon=1e-5, \
    epochs=100):
        Trains network until it converge. Parameter ``epochs`` control
        maximum number of iterations, just to make sure that network will
        stop training procedure if it can't converge.
    """,
    "supervised_train_lazy": """train(input_train, target_train, copy=True):
        Network just stores all the information about the data and use it for \
        the prediction. Parameter ``copy`` copy input data before store it \
        inside the network.
    """,

    # ------------------------------------ #
    #              Parameters              #
    # ------------------------------------ #

    "verbose": """verbose : bool
        Property controls verbose output interminal.
        ``True`` enable informative output in the terminal and ``False`` -
        disable it. Defaults to ``False``.
    """,
    "step": """step : float
        Learns step, defaults to ``0.1``.
    """,
    "show_epoch": """show_epoch : int or str
        This property controls how often the network will display information
        about training. There are two main syntaxes for this property.
        You can describe it as positive integer number and it
        will describe how offen would you like to see summary output in
        terminal. For instance, number `100` mean that network will show you
        summary in 100, 200, 300 ... epochs. String value should be in a
        specific format. It should contain the number of times that the output
        will be displayed in the terminal. The second part is just
        a syntax word ``time`` or ``times`` just to make text readable.
        For instance, value ``'2 times'`` mean that the network will show
        output twice with approximately equal period of epochs and one
        additional output would be after the finall epoch.
        Defaults to ``'10 times'``.
    """,
    "shuffle_data": """shuffle_data : bool
        If it's ``True`` class shuffles all your training data before
        training your network, defaults to ``True``.
    """,
    "error": """error : function
        Function which controls your training error. defaults to ``mse``
    """,
    "use_bias": """use_bias : bool
        Uses bias in the network, defualts to ``True``.
    """,
    "train_epoch_end_signal": """train_epoch_end_signal : function
        Calls this function when train epoch finishes.
    """,
    "train_end_signal": """train_end_signal : function
        Calls this function when train process finishes.
    """,

    # ------------------------------------ #
    #                 Steps                #
    # ------------------------------------ #
    "first_step": """first_step : float
        Contains initialized step value.
    """,
    "steps": """steps : list of float
        List of steps in the same order as the network layers.
        By default all values are equal to ``step`` parameter.
    """,

    # ------------------------------------ #
    #               Warnings               #
    # ------------------------------------ #
    "bp_depending": """It works with any algorithm based on backpropagation. \
    Class can't work without it.
    """
}


# ------------------------------------ #
#         Complex parameters           #
# ------------------------------------ #


def joindocs(docs, docskeys):
    return ''.join([docs[key] for key in docskeys])


full_params_params = ()
docs.update({
    'full_params': joindocs(
        docs,
        [
            'step', 'show_epoch', 'shuffle_data',
            'error', 'use_bias', 'train_epoch_end_signal',
            'train_end_signal', 'verbose'
        ]
    ),
    'full_signals': joindocs(
        docs, ['train_epoch_end_signal', 'train_end_signal']
    ),
    'full_methods': joindocs(
        docs, ['fit', 'predict', 'last_error', 'plot_errors']
    )
})



import sys

import numpy as np


__all__ = ('format_data', 'is_row1d')


def format_data(input_data, row1d=False, copy=False):
    if input_data is None:
        return

    input_data = np.array(input_data, copy=copy)

    # Valid number of features for one or two dimentions
    n_features = input_data.shape[-1]
    if 'pandas' in sys.modules:
        pandas = sys.modules['pandas']

        if isinstance(input_data, (pandas.Series, pandas.DataFrame)):
            input_data = input_data.values
    if input_data.ndim == 1:
        data_shape = (1, n_features) if row1d else (n_features, 1)
        input_data = input_data.reshape(data_shape)

    return input_data


def is_row1d(layer):
    if layer is None:
        return False
    return (layer.input_size != 1)



import numpy as np


__all__ = ('add_bias_column', 'iter_until_converge', 'shuffle')


def add_bias_column(data):
    """ Transform input data to matrix and add as first column which
    contains all ones.

    Parameters
    ----------
    data : array-like
        Main matrix.

    Returns
    -------
    array-like
        Copied matrix with the new column at first position which contains
        all ones.
    """
    data = np.asmatrix(data)
    bias_vector = np.ones((data.shape[0], 1))
    return np.concatenate((bias_vector, np.asarray(data)), axis=1)


def iter_until_converge(network, epsilon, max_epochs):
    # Trigger first iteration and store first error term
    yield network.epoch
    error = network.last_error()

    while error > epsilon:
        network.epoch += 1
        yield network.epoch
        error = abs(network.last_error() - error)

        if network.epoch >= max_epochs and error > epsilon:
            network.logs.log(
                "TRAIN",
                "Epoch #{} stopped. Network didn't converge "
                "after {} iterations".format(network.epoch, max_epochs)
            )
            return

    if np.isnan(error) or np.isinf(error):
        network.logs.log("TRAIN", "Epoch #{} stopped. Network error value is "
                                  "invalid".format(network.epoch))
    else:
        network.logs.log("TRAIN", "Epoch #{} stopped. Network converged."
                                  "".format(network.epoch))


def shuffle(*arrays):
    if not arrays:
        return tuple()

    first = arrays[0]
    n_samples = first.shape[0]

    for array in arrays:
        if n_samples != array.shape[0]:
            raise ValueError("All matrices muts have the same number of rows")

    indices = np.arange(n_samples)
    np.random.shuffle(indices)

    arrays = list(arrays)
    for i, array in enumerate(arrays):
        arrays[i] = array[indices]

    return arrays
